; -------------------------- RANROT32.ASM -------------------------------
; random number generator by Agner Fog 1999
; 32-bit mode version for 80x86 and compatible microprocessors.
; Works with Window 95 and later, and possibly with other 32-bit systems.
;
; The code uses the RANROT type W algorithm with self-test.
; The resolution is 63 bits for WRandom and 32 bits for WBRandom.
;
; WRandomInit must be called before the first call to any of the random
; functions.
; WRandom returns a floating point number between 0 and 1.
; WBRandom returns an integer between 0 and 0xFFFFFFFF
; WIRandom returns an integer in the interval set by WSetInterval
;
; C++ prototypes:
; extern"C" void _cdecl WRandomInit (int seed);
; extern"C" void _cdecl WSetInterval (int min, int max);
; extern"C" double _cdecl WRandom (void);
; extern"C" unsigned _cdecl WBRandom (void);
; extern"C" int _cdecl WIRandom (void);
;
; -----------------------------------------------------------------------

; The random function is optimized for the Pentium microprocessor.

.386

; define appropriate function for error-exit:
; (depends on system):
; ErrorExit equ <FatalAppExitA>

; define parameters
JJ      equ     10            ; lag 1
KK      equ     17            ; lag 2, size of circular buffer
R1      equ     19            ; rotate count
R2      equ     27            ; rotate count
SAFE    equ      1            ; define this if you want self-test

RAND_DATA SEGMENT PARA USE32 PUBLIC 'DATA'
align   16
randp1  DT      1.5           ; used for conversion to float
        DW      0             ; alignment
p1      DD      0             ; pointer in circular buffer
p2      DD      0             ; pointer in circular buffer
randbuf DD      (2*KK) dup(?) ; circular buffer
one     DD      1.0           ; constant
min     DD      0             ; used by WSetInterval
interval DD     0             ; used by WSetInterval
ALIGN   8
finterval DQ    0             ; used by WSetInterval

IFDEF   SAFE
randbufcopy DD  (4*KK) dup(?) ; double copy of randbuf initial state
RandMessage1    DB 'Random number generator not initialized',0
RandMessage2    DB 'Random number generator returned to initial state',0
ENDIF
RAND_DATA ENDS


_TEXT   SEGMENT DWORD USE32 PUBLIC 'CODE'
FLAT GROUP RAND_DATA,_TEXT
ASSUME CS:FLAT,DS:FLAT,ES:FLAT

PUBLIC WRandomInit,_WRandomInit,WSetInterval,_WSetInterval
PUBLIC WRandom,_WRandom,WBRandom,_WBRandom,WIRandom,_WIRandom

IFNDEF  SAFE                    ; without self-test
WRandom PROC NEAR
_WRandom LABEL NEAR
        FLD     [randp1]        ; load random number
        FSUB    [one]           ; normalize
Random3 LABEL NEAR              ; second entry for integer output
        MOV     ECX, [p1]       ; ring buffer pointer
        MOV     EDX, [p2]       ; ring buffer pointer
        PUSH    EBX
        NOP                     ; avoid imperfect pairing due to AGI
        MOV     EBX, [randbuf][ECX]
        MOV     EAX, [randbuf][ECX+4]

ELSE                            ; include self-test
Random3 LABEL NEAR              ; second entry for integer output
        MOV     ECX, [p1]       ; ring buffer pointers
        MOV     EAX, [randbufcopy][0] ; first dword of initial state
        JMP     SHORT R10
WRandom PROC NEAR
_WRandom LABEL NEAR
        FLD     [randp1]        ; load random number
        MOV     ECX, [p1]       ; ring buffer pointers
        MOV     EAX, [randbufcopy][0] ; first dword of initial state
        FSUB    [one]           ; normalize random number
R10:    PUSH    EBX
        MOV     EBX, [randbuf][ECX]
        CMP     EAX, EBX        ; self-test first dword
        JE      SHORT R50       ; jump if further compare needed
R20:    MOV     EDX, [p2]       ; ring buffer pointer
        MOV     EAX, [randbuf][ECX+4]
ENDIF
        ROL     EBX, R1                  ; rotate bits
        ROL     EAX, R2
        ADD     EBX, [randbuf][EDX]      ; add two dwords
        ADD     EAX, [randbuf][EDX+4]
        MOV     [randbuf][ECX], EAX      ; save in swapped order
        MOV     [randbuf][ECX+4], EBX
        OR      EBX, 80000000H           ; bit 63 must be set
        MOV     DWORD PTR [randp1],EAX   ; low 32 bits of next random number
        SUB     ECX, 8                   ; decrement p1
        JNC     SHORT R30
        MOV     ECX, (KK-1)*8            ; wrap around p1
R30:    MOV     DWORD PTR [randp1+4],EBX ; bits 32-62 of next random number
        POP     EBX
        SUB     EDX, 8                   ; decrement p2
        JNC     SHORT R40
        MOV     EDX, (KK-1)*8            ; wrap around p2
R40:    MOV     [p1], ECX                ; save updated pointers
        MOV     [p2], EDX
        RET
IFDEF   SAFE
R50:    ; first dword is identical. compare entire buffer
        PUSHAD
        MOV     ESI, OFFSET DS:randbuf
        MOV     EDI, OFFSET DS:randbufcopy+KK*8
        SUB     EDI, ECX                  ; buffer is rotated
        MOV     ECX, KK*2
        CLD
        REPE    CMPSD                     ; compare buffers
        JE      SHORT R60
        POPAD
        JMP     R20                       ; buffers are different. continue
R60:    ; error found. initial state is repeated
IFDEF   ErrorExit
        PUSH    OFFSET DS:RandMessage1
        MOV     EAX, [p1]
        CMP     EAX, [p2]
        JE      SHORT R70
        PUSH    OFFSET DS:RandMessage2
R70:    PUSH    0
        CALL    ErrorExit   ; name of abort procedure depends on system
        EXTRN   ErrorExit:NEAR
ELSE
        ; ErrorExit function not defined. Make division by 0 error
        DIV     ECX
ENDIF
ENDIF
WRandom ENDP

WBRandom PROC NEAR
_WBRandom LABEL NEAR
        MOV     EAX, DWORD PTR [randp1]
        PUSH    EAX
        CALL    Random3
        POP     EAX
        RET
WBRandom ENDP

WSetInterval PROC NEAR
_WSetInterval LABEL NEAR
        MOV     EAX, [ESP+4]            ; min
        MOV     EDX, [ESP+8]            ; max
        CMP     EAX, EDX
        JL      RSI1
        XCHG    EAX, EDX                ; make sure min <= max
RSI1:   MOV     [min], EAX
        SUB     EDX, EAX
        INC     EDX
        MOV     [interval], EDX         ; max-min
        FILD    [interval]
        FSTP    [finterval]
        RET
WSetInterval ENDP

WIRandom PROC   NEAR
_WIRandom LABEL NEAR
        CALL    WRandom
        FMUL    [finterval]
        SUB     ESP, 4
        FISTP   DWORD PTR [ESP]
      ; WAIT                    ;(only needed for old processors)
        POP     EAX
        CMP     EAX,[interval]
        ADC     EAX,-1          ; make sure value < interval
        ADD     EAX,[min]
        RET
WIRandom ENDP

WRandomInit PROC NEAR
_WRandomInit LABEL NEAR
        MOV     EAX, [ESP+4]    ; seed
        CMP     EAX, 1
        SBB     EAX, 0          ; make sure seed != 0
        XOR     ECX, ECX
        ; make random numbers and put them into buffer
R80:    MOV     EDX, EAX
        SHL     EAX, 13
        XOR     EDX, EAX
        MOV     EAX, EDX
        SHR     EDX, 17
        XOR     EAX, EDX
        MOV     EDX, EAX
        SHL     EDX, 5
        XOR     EAX, EDX
        MOV     [randbuf][ECX*4], EAX
        INC     ECX
        CMP     ECX, KK*2
        JB      R80
        FLD1
        FSTP    [randp1]        ; initialize bit 64-79
        MOV     [p1], 0         ; initialize buffer pointers
        MOV     [p2], JJ*8
IFDEF   SAFE
        PUSHAD                  ; make 2 copies of initial state
        MOV     ECX, KK*2
        MOV     ESI, OFFSET DS:randbuf
        MOV     EDI, OFFSET DS:randbufcopy
        PUSH    ECX
        PUSH    ESI
        CLD
        REP     MOVSD
        POP     ESI
        POP     ECX
        REP     MOVSD
        INC     [randbufcopy][0]   ; avoid error message
        CALL    Random3            ; prepare first random number
        DEC     [randbufcopy][0]
        POPAD
ELSE
        CALL    Random3            ; prepare first random number
ENDIF
        PUSH    EDI
        MOV     EDI, 96
R90:    CALL    Random3
        DEC     EDI
        JNZ     R90
        POP     EDI                                                

        RET     0                  ; RET 4 if not _cdecl calling
WRandomInit ENDP

_TEXT   ENDS
        END

; To optimize for Pentium II processor: Remove the NOP and replace
; the conditional jumps to R30 and R40 with conditional moves.


; Example of calling from C++ program:

; #include "arandom.h"
; int i;
; WRandomInit (time(0)); WSetInterval(1,10);
; for (i=0; i<100; i++) {
;   printf ("\n%14.10f  %2i", WRandom(), WIRandom);}
